// ----------------------------------------------------------------------^
// Copyright (C) 2004, 2005, 2006, 2007, 2008 Giorgio Calderone
// (mailto: <gcalderone@ifc.inaf.it>)
// 
// This file is part of MCS.
// 
// MCS is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// MCS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with MCS; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// 
// ----------------------------------------------------------------------$
#include <mcs.hh>
using namespace mcs;

#define TEST_USER "mcstest"
#define TEST_PASS "mcstest"
#define TEST_DBNA "test"
#define TEST_TBL  "mcstest"

/* Trim leading and trailing quote (') and spaces */
std::string trimfs( std::string s ) {
  string ss = s;
  ss.erase( 0, ss.find_first_not_of( " '\t\n" ) );
  ss.erase( ss.find_last_not_of( " '\t\n" ) + 1);
  return ss;
}

/* The file name without the dir. path and the extension (e.g. ".fits") */
std::string fclean( std::string fullname, std::string fext ) {
  string fnam;
  size_t sl = fullname.rfind("/");
  if (sl != string::npos)
    fnam = fullname.substr(sl+1);
  else
    fnam = fullname;

  sl = fnam.find(fext);

 if (sl != string::npos)
   return fnam.erase(sl);
 else
   return fnam;
}


int main(int argc, char* argv[])
{
  FITSReader fits;
  int i, j, hdu, nflds=0;

  DBConn db;
  Query* qry;


  try {
    for (i=1; i<argc; i++) {
      cout  << endl  << endl << "Reading: " << argv[i] << endl;

      fits.open(argv[i]);

      for (hdu=1; hdu<=fits.HDUCount(); hdu++) {
	fits.selectHDU(hdu);
	cout << endl << "HDU #" << fits.currentHDU() << endl;

	//Print header
	for (j=0; j<fits.header.count(); j++) {
	  cout <<
	    fits.header[j].name() << " = " <<
	    fits.header[j].sval() << " / " <<
	    fits.header_comments[j].sval() << endl;
	}

        nflds = fits.metarec().count();
cout << "There are "<< nflds << " fields."<<endl;
	//Print field names
	cout << fits.metarec().asStringNames() << endl;

	//Print field types
	cout << fits.metarec().asStringTypes() << endl;

	//Print data
/*
	while (! fits.eof()) {
	  cout << fits.rec().asString() << endl;
	  fits.setNext();
	}
*/
      }


// Only create table for the last HDU (can be changed)

cout <<"There are "<< fits.header["NAXIS2"].sval() <<" rows"<< endl;;

      string kname, ktype, flds;
      string tabn = fclean(string(argv[i]), ".");
      Types type;
//enum enum_field_types mtype;

      kname = fits.metarec().field(0).name();
      flds = "(`"+ kname +"` ";

      if (fits.metarec().field(0).type() == STRING)
        ktype = "VARCHAR("+ itos(fits.metarec().field(0).length()) +")";
      else {
        //ktype = Types2Str(fits.metarec().field(0).type(), fits.metarec().field(0).isUnsigned());
          type = fits.metarec().field(0).type();
          //if (Types2MYSQL(type, mtype))
            //ktype = MYSQL2Str(mtype);
ktype = Types2MYSQLStr(type, fits.metarec().field(0).isUnsigned());
      }

      flds += ktype;

      for (j=1; j<fits.metarec().count(); j++) {
        kname = fits.metarec().field(j).name();
        flds += ", `"+ kname +"` ";
//cout <<j<<" "<<kname<<" "<<ktype<<"**"<< endl;
        if (fits.metarec().field(j).type() == STRING)
          ktype = "VARCHAR("+ itos(fits.metarec().field(j).length()) +")";
        else {
          //ktype = Types2Str(fits.metarec().field(j).type(), fits.metarec().field(j).isUnsigned());
          type = fits.metarec().field(j).type();
          //if (Types2MYSQL(type, mtype))
            //ktype = MYSQL2Str(mtype);
ktype = Types2MYSQLStr(type, fits.metarec().field(j).isUnsigned());
        }
        flds += ktype;
      }

      flds += ")";

// To DB table
      db.connect(TEST_USER, TEST_PASS, TEST_DBNA, "localhost");
      qry = new Query(&db);
      qry->query("CREATE TABLE IF NOT EXISTS "+ tabn +" "+ flds);
      qry->prepare_with_parameters(MCS_PAB_INSERT, "*", tabn);


// Insert the data
      while (! fits.eof()) {
//cout << fits.rec().asString() << endl;
        for (j=0; j<nflds; j++)
          qry->param()[j] = fits.rec()[j].sval();
        fits.setNext();

        qry->execute();
      }

      delete qry;
      db.close();


      fits.close();
    }
  }
  catch(Event& e) {
    cerr << e.msg() << endl;
  }
}
